/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Gradient
Version:   1.0
ID:        ID_GRADIENT
Status:    Stable
Category:  GUI
Date:      August 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Keywords:  gradient, graphics, colours, colour, blend, rainbow
Short:     The gradient class draws a range of gradually changing colours in
           a chosen direction.

###DESCRIPTION###
<p>The Gradient class is an excellent tool for creating visually appealing
backdrops for your graphics.  Although the concept is quite simple, gradients
can be used to simulate everything from images of the sky, fire, metals,
water, horizons, or sun-sets.</p>

<p>To create a gradient you will need to decide on just a few factors - the
colour to start the gradient, the colour to end it, the direction of the
gradient, and if necessary the positioning and dimensions of the graphical
area.  A gradient may be drawn in four different directions - either
horizontally, vertically, from the top left to the bottom right corner, or
the bottom left to the top right corner.  Specific angles from 0 to 360
degrees are not yet supported.</p>

<p>The following examples illustrate common usage of the Gradient class:</p>

<pre>
<i>Horizontal Gradient, White to Black:</i>

  &lt;gradient direction="horziontal" firstcolour="#ffffff" lastcolour="#000000"/&gt;

<i>Vertical Gradient, Red to Blue, 200x200:</i>

  &lt;gradient direction="vertical" firstcolour="#ff0000" lastcolour="#0000ff"
    x="100" y="100" width="200" height="200"/&gt;
</pre>

<p>All standard GUI constructs are supported, so percentage based dimensions
and offsets can also be used.  In cases where no coordinates or dimensions
have been specified, the gradient will take up the entire graphical area of
its related drawable.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/functions/strtoint.c>
#include <math.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_GRADIENT, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003.  All rights reserved.",
   "August 2003",
   "Gradient"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR GradientClass = 0;
static UBYTE *glAlphaLookup;

#define INTERP_COSINE true

static UBYTE glDither8[8][8] = {
   {  0, 32,  8, 40,  2, 34, 10, 42 },
   { 48, 16, 56, 24, 50, 18, 58, 26 },
   { 12, 44,  4, 36, 14, 46,  6, 38 },
   { 60, 28, 52, 20, 62, 30, 54, 22 },
   {  3, 35, 11, 43,  1, 33,  9, 41 },
   { 51, 19, 59, 27, 49, 17, 57, 25 },
   { 15, 47,  7, 39, 13, 45,  5, 37 },
   { 63, 31, 55, 23, 61, 29, 53, 21 }
};

/****************************************************************************
** Gradient class.
*/

struct Gradient {
   OBJECT_HEADER
   LONG Direction; /* Direction of the gradient (vertical, horizontal, topdiagonal, bottomdiagonal) */
   LONG Frame;

   PRIVATE_FIELDS
   OBJECTID DrawableID;
   struct RGB FirstColourRGB;
   struct RGB LastColourRGB;
   LONG  Flags;
   LONG  Dimensions;
   LONG  Opacity;
   LONG  BitsPerPixel;
   DOUBLE Height;
   DOUBLE Width;
   DOUBLE XCoord;
   DOUBLE XOffset;
   DOUBLE YCoord;
   DOUBLE YOffset;
   WORD *Frames;
   WORD TotalFrames;
   ULONG Colours[256];   /* Pregenerated colour lookup table */
};

static ERROR GET_FirstColourRGB(struct Gradient *, struct RGB *);
static ERROR GET_LastColourRGB(struct Gradient *, struct RGB *);
static ERROR GET_Height(struct Gradient *, struct Variable *);
static ERROR GET_Opacity(struct Gradient *, FLOAT *);
static ERROR GET_Width(struct Gradient *, struct Variable *);
static ERROR GET_X(struct Gradient *, struct Variable *);
static ERROR GET_XOffset(struct Gradient *, struct Variable *);
static ERROR GET_Y(struct Gradient *, struct Variable *);
static ERROR GET_YOffset(struct Gradient *, struct Variable *);

static ERROR SET_FirstColour(struct Gradient *, STRING);
static ERROR SET_FirstColourRGB(struct Gradient *Self, struct RGB *);
static ERROR SET_Frames(struct Gradient *, STRING);
static ERROR SET_LastColour(struct Gradient *, STRING);
static ERROR SET_LastColourRGB(struct Gradient *Self, struct RGB *);
static ERROR SET_Height(struct Gradient *, struct Variable *);
static ERROR SET_Opacity(struct Gradient *, FLOAT *);
static ERROR SET_Width(struct Gradient *, struct Variable *);
static ERROR SET_X(struct Gradient *, struct Variable *);
static ERROR SET_XOffset(struct Gradient *, struct Variable *);
static ERROR SET_Y(struct Gradient *, struct Variable *);
static ERROR SET_YOffset(struct Gradient *, struct Variable *);

/*** General flags ***/

#define GRF_VISIBLE 0x00000001

static struct FieldDef DirectionTable[] = {
   { "VERTICAL",       GDI_VERTICAL },
   { "HORIZONTAL",     GDI_HORIZONTAL },
   { "TOPDIAGONAL",    GDI_TOPDIAGONAL },
   { "BOTTOMDIAGONAL", GDI_BOTTOMDIAGONAL },
   { "DIAGONAL",       GDI_TOPDIAGONAL },
   { NULL, NULL }
};

static struct FieldArray GradientFields[] = {
   { "Direction",      0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&DirectionTable, NULL, NULL },
   { "Frame",          0, FDF_LONG|FDF_RW,            0, NULL, NULL },
   /*** Virtual Fields ***/
   { "FirstColour",    0, FDF_STRING|FDF_W,  0, NULL,               SET_FirstColour },
   { "FirstColourRGB", 0, FDF_POINTER|FDF_R, 0, GET_FirstColourRGB, SET_FirstColourRGB },
   { "Frames",         0, FDF_STRING|FDF_W,  0, NULL,               SET_Frames },
   { "LastColour",     0, FDF_STRING|FDF_W,  0, NULL,               SET_LastColour },
   { "LastColourRGB",  0, FDF_POINTER|FDF_R, 0, GET_LastColourRGB,  SET_LastColourRGB },
   { "Opacity",        0, FDF_FLOAT|FDF_RW,  0, GET_Opacity,        SET_Opacity },
   /*** Variable Fields ***/
   { "Height",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",          0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Gradient actions
*/

static ERROR GRADIENT_Free(struct Gradient *, APTR);
static ERROR GRADIENT_Hide(struct Gradient *, APTR);
static ERROR GRADIENT_Init(struct Gradient *, APTR);
static ERROR GRADIENT_NewObject(struct Gradient *, APTR);
static ERROR GRADIENT_Show(struct Gradient *, APTR);

static struct ActionArray GradientActions[] = {
   { AC_Free,         GRADIENT_Free },
   { AC_Hide,         GRADIENT_Hide },
   { AC_Init,         GRADIENT_Init },
   { AC_NewObject,    GRADIENT_NewObject },
   { AC_Show,         GRADIENT_Show },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Bitmap module (to create the alpha blending table) ***/

   if (AccessMemory(RPM_AlphaBlend, MEM_READWRITE|MEM_NOBLOCKING, &glAlphaLookup) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_AccessMemory));
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&GradientClass, NULL,
      FID_BaseClassID|TLONG, ID_GRADIENT,
      FID_Version|TFLOAT,    VER_GRADIENT,
      FID_Name|TSTRING,      "Gradient",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      GradientActions,
      FID_Fields|TPTR,       GradientFields,
      FID_Flags|TLONG,       CLF_PRIVATEONLY,
      FID_Size|TLONG,        sizeof(struct Gradient),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (glAlphaLookup) { ReleaseMemory(glAlphaLookup); glAlphaLookup = NULL; }

   if (GradientClass) { Action(AC_Free, GradientClass, NULL); GradientClass = 0; }
   return(ERR_Okay);
}

/******************************************************************************
** Internal: CalculateColours()
*/

static void CalculateColours(struct Gradient *Self, struct Bitmap *Bitmap)
{
#ifdef INTERP_COSINE
   struct ColourFormat *format;
   WORD i;
   UBYTE red, green, blue, alpha;
   FLOAT theta, f;

   /* We use the cosine interpoloation method for the generating the
   ** gradient colour scheme.  This produces the best quality colour
   ** transition.
   */

   format = Bitmap->ColourFormat;
   for (i=0; i < ARRAYSIZE(Self->Colours); i++) {
      theta = ((FLOAT)i / 255.0) * 3.1415927;
      f     = (1 - cos(theta)) * 0.5;
      red   = (((FLOAT)Self->FirstColourRGB.Red)   * (1-f) + ((FLOAT)Self->LastColourRGB.Red) * f);
      green = (((FLOAT)Self->FirstColourRGB.Green) * (1-f) + ((FLOAT)Self->LastColourRGB.Green) * f);
      blue  = (((FLOAT)Self->FirstColourRGB.Blue)  * (1-f) + ((FLOAT)Self->LastColourRGB.Blue) * f);

      if (Self->Opacity < 255) alpha = Self->Opacity;
      else alpha = (((FLOAT)Self->FirstColourRGB.Alpha) * (1-f) + ((FLOAT)Self->LastColourRGB.Alpha) * f);

      Self->Colours[i] = (((red>>format->RedShift)     & format->RedMask)   << format->RedPos) |
                         (((green>>format->GreenShift) & format->GreenMask) << format->GreenPos) |
                         (((blue>>format->BlueShift)   & format->BlueMask)  << format->BluePos) |
                         (alpha << 24);
   }

#else

   /* Linear interpolation is simple and fast, but the quality
   ** is not as high as the cosine method.
   */

   WORD reddist, greendist, bluedist, alphadist;

   reddist   = Self->LastColourRGB.Red   - Self->FirstColourRGB.Red;
   greendist = Self->LastColourRGB.Green - Self->FirstColourRGB.Green;
   bluedist  = Self->LastColourRGB.Blue  - Self->FirstColourRGB.Blue;
   alphadist = Self->LastColourRGB.Alpha - Self->FirstColourRGB.Alpha;

   for (i=0; i < ARRAYSIZE(Self->Colours); i++) {
      Self->Colours[i].Alpha = (Self->FirstColourRGB.Alpha + ((alphadist * i)>>8));
      Self->Colours[i].Red   = (Self->FirstColourRGB.Red   + ((reddist   * i)>>8)) & redmask;
      Self->Colours[i].Green = (Self->FirstColourRGB.Green + ((greendist * i)>>8)) & greenmask;
      Self->Colours[i].Blue  = (Self->FirstColourRGB.Blue  + ((bluedist  * i)>>8)) & bluemask;
   }

#endif
}

/*****************************************************************************
** Internal: DrawGradient()
*/

static void DrawGradient(struct Gradient *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct Variable var;
   struct mtDrawRGBLine line;
   struct RGB rgb;
   WORD cindex, nindex, xindex;
   ULONG colour, nextcolour, dcolour;
   LONG i, sx, sy, x, y, value, width, height, pos;
   LONG RedDist, GreenDist, BlueDist, AlphaDist, opacity;
   WORD current, next, clipright, clipleft, cliptop, clipbottom;
   UBYTE *srctable, *desttable;
   UBYTE dither;

   if (Self->BitsPerPixel != Bitmap->BitsPerPixel) {
      CalculateColours(Self, Bitmap);
      Self->BitsPerPixel = Bitmap->BitsPerPixel;
   }

   /* Determine whether or not we need to draw this gradient
   ** based on our object's frame settings.
   */

   if (Self->Frame) {
      if (Drawable->Frame != Self->Frame) return;
   }
   else if (Self->Frames) {
      if ((Drawable->Frame > Self->TotalFrames) OR (Self->Frames[Drawable->Frame] < 1)) return;
   }

   /*** Get the object dimensions ***/

   var.Type = FD_LARGE;
   GET_X(Self, &var); x = var.Large;
   GET_Y(Self, &var); y = var.Large;
   GET_Width(Self, &var);  width = var.Large;
   GET_Height(Self, &var); height = var.Large;

   if ((width <= 0 ) OR (height <= 0)) return;

   if (Self->Direction IS GDI_HORIZONTAL) {

      /*** Clip the vertical coordinates first ***/

      if (y < Bitmap->ClipTop) {
         height = height - (Bitmap->ClipTop - y);
         y = Bitmap->ClipTop;
      }

      if (y + height >= Bitmap->ClipBottom) {
         height = Bitmap->ClipBottom - y;
      }

      x += Bitmap->XOffset;
      y += Bitmap->YOffset;
      clipleft  = Bitmap->ClipLeft  + Bitmap->XOffset;
      clipright = Bitmap->ClipRight + Bitmap->XOffset;
      current = 0;
      next    = 1;
      colour  = ((ULONG *)Self->Colours)[0] + 1;
      nextcolour = ((ULONG *)Self->Colours)[0];

      for (sx=x, value=0; value < width; sx++, value++, current++) {
         if (sx >= clipright) break;

         cindex = (value<<8) / width;
         if (Self->Colours[cindex] != colour) {
            colour     = Self->Colours[cindex];
            nextcolour = Self->Colours[cindex];
            current = 0;
            for (next=1; value+next < width; next++) {
               nindex = ((value+next)<<8) / width;
               if (((ULONG *)Self->Colours)[nindex] != colour) {
                  nextcolour = Self->Colours[nindex];
                  break;
               }
            }
         }

         if (sx < clipleft) continue;

         xindex = ((current<<6) / next);
         dither = sx & 0x7;
         for (sy=y; sy < y + height; sy++) {
            if (glDither8[sy & 0x7][dither] < xindex) dcolour = nextcolour;
            else dcolour = colour;

            if ((dcolour>>24) < 255) {
               srctable  = glAlphaLookup + ((dcolour>>24)<<8);
               desttable = glAlphaLookup + ((255 - (dcolour>>24))<<8);
               Bitmap->ReadUCRPixel(Bitmap, sx, sy, &rgb);
               dcolour = PackPixel(Bitmap, srctable[UnpackRed(Bitmap,dcolour)]   + desttable[rgb.Red],
                                           srctable[UnpackGreen(Bitmap,dcolour)] + desttable[rgb.Green],
                                           srctable[UnpackBlue(Bitmap,dcolour)]  + desttable[rgb.Blue]);
               Bitmap->DrawUCPixel(Bitmap, sx, sy, dcolour);
            }
            else Bitmap->DrawUCPixel(Bitmap, sx, sy, dcolour & 0xffffff);
         }
      }
   }
   else if (Self->Direction IS GDI_VERTICAL) {
      /*** Clip the horizontal coordinates first ***/

      if (x < Bitmap->ClipLeft) {
         width = width - (Bitmap->ClipLeft - x);
         x = Bitmap->ClipLeft;
      }

      if (x + width >= Bitmap->ClipRight) {
         width = Bitmap->ClipRight - x;
      }

      x += Bitmap->XOffset;
      y += Bitmap->YOffset;
      cliptop    = Bitmap->ClipTop + Bitmap->YOffset;
      clipbottom = Bitmap->ClipBottom + Bitmap->YOffset;
      current = 0;
      next    = 1;
      colour  = ((ULONG *)Self->Colours)[0] + 1;
      nextcolour = ((ULONG *)Self->Colours)[0];

      for (sy=y, value=0; value < height; sy++, value++, current++) {
         if (sy >= clipbottom) break;

         cindex = (value<<8) / height;
         if (Self->Colours[cindex] != colour) {
            colour     = Self->Colours[cindex];
            nextcolour = Self->Colours[cindex];
            current = 0;
            for (next=1; value+next < height; next++) {
               nindex = ((value+next)<<8) / height;
               if (((ULONG *)Self->Colours)[nindex] != colour) {
                  nextcolour = Self->Colours[nindex];
                  break;
               }
            }
         }

         if (sy < cliptop) continue;

         xindex = ((current<<6) / next);
         dither = sy & 0x7;
         for (sx=x; sx < x + width; sx++) {
            if (glDither8[dither][sx & 0x7] < xindex) dcolour = nextcolour;
            else dcolour = colour;

            if ((dcolour>>24) < 255) {
               srctable  = glAlphaLookup + ((dcolour>>24)<<8);
               desttable = glAlphaLookup + ((255 - (dcolour>>24))<<8);
               Bitmap->ReadUCRPixel(Bitmap, sx, sy, &rgb);
               dcolour = PackPixel(Bitmap, srctable[UnpackRed(Bitmap,dcolour)]   + desttable[rgb.Red],
                                           srctable[UnpackGreen(Bitmap,dcolour)] + desttable[rgb.Green],
                                           srctable[UnpackBlue(Bitmap,dcolour)]  + desttable[rgb.Blue]);
               Bitmap->DrawUCPixel(Bitmap, sx, sy, dcolour);
            }
            else Bitmap->DrawUCPixel(Bitmap, sx, sy, dcolour & 0xffffff);
         }
      }
   }
   else if (Self->Direction IS GDI_TOPDIAGONAL) {
      /*** Diagonal (from top left) ***/

      opacity = Bitmap->Opacity;
      Bitmap->Opacity = Self->Opacity;

      RedDist   = Self->LastColourRGB.Red   - Self->FirstColourRGB.Red;
      GreenDist = Self->LastColourRGB.Green - Self->FirstColourRGB.Green;
      BlueDist  = Self->LastColourRGB.Blue  - Self->FirstColourRGB.Blue;
      AlphaDist = Self->LastColourRGB.Alpha - Self->FirstColourRGB.Alpha;

      line.XCoord = x;
      line.XEnd   = x;
      line.YCoord = y;
      line.YEnd   = y;
      line.Red    = Self->FirstColourRGB.Red;
      line.Green  = Self->FirstColourRGB.Green;
      line.Blue   = Self->FirstColourRGB.Blue;
      if (Self->Opacity IS 255) Bitmap->Opacity = Self->FirstColourRGB.Alpha;

      if (RedDist)   RedDist   = (1024 * (width + height)) / RedDist;
      if (GreenDist) GreenDist = (1024 * (width + height)) / GreenDist;
      if (BlueDist)  BlueDist  = (1024 * (width + height)) / BlueDist;

      pos = 0;
      for (i=0; i < width + height - 1; i++) {
         if (RedDist)   line.Red   = Self->FirstColourRGB.Red   + (pos / RedDist);
         if (GreenDist) line.Green = Self->FirstColourRGB.Green + (pos / GreenDist);
         if (BlueDist)  line.Blue  = Self->FirstColourRGB.Blue  + (pos / BlueDist);
         if ((AlphaDist) AND (Self->Opacity IS 255))  Bitmap->Opacity = Self->FirstColourRGB.Alpha + (pos / AlphaDist);
         Action(MT_DrawRGBLine, Bitmap, &line);
         pos += 1024;

         if (line.YCoord IS (y + height - 1)) line.XCoord++;
         else line.YCoord++;

         if (line.XEnd IS (x + width - 1)) line.YEnd++;
         else line.XEnd++;
      }

      Bitmap->Opacity = opacity;
   }
   else if (Self->Direction IS GDI_BOTTOMDIAGONAL) {
      /*** Diagonal (from bottom left) ***/

      opacity = Bitmap->Opacity;
      Bitmap->Opacity = Self->Opacity;

      RedDist   = Self->LastColourRGB.Red   - Self->FirstColourRGB.Red;
      GreenDist = Self->LastColourRGB.Green - Self->FirstColourRGB.Green;
      BlueDist  = Self->LastColourRGB.Blue  - Self->FirstColourRGB.Blue;
      AlphaDist = Self->LastColourRGB.Alpha - Self->FirstColourRGB.Alpha;

      line.XCoord = x;
      line.XEnd   = x;
      line.YCoord = y + height - 1;
      line.YEnd   = y + height - 1;
      line.Red    = Self->FirstColourRGB.Red;
      line.Green  = Self->FirstColourRGB.Green;
      line.Blue   = Self->FirstColourRGB.Blue;
      if (Self->Opacity IS 255) Bitmap->Opacity = Self->FirstColourRGB.Alpha;

      if (RedDist)   RedDist   = (1024 * (width + height)) / RedDist;
      if (GreenDist) GreenDist = (1024 * (width + height)) / GreenDist;
      if (BlueDist)  BlueDist  = (1024 * (width + height)) / BlueDist;
      if ((AlphaDist) AND (Self->Opacity IS 255)) Bitmap->Opacity = (1024 * (width + height)) / AlphaDist;

      pos = 0;
      for (i=0; i < width + height - 1; i++) {
         if (RedDist)   line.Red   = Self->FirstColourRGB.Red   + (pos / RedDist);
         if (GreenDist) line.Green = Self->FirstColourRGB.Green + (pos / GreenDist);
         if (BlueDist)  line.Blue  = Self->FirstColourRGB.Blue  + (pos / BlueDist);
         if ((AlphaDist) AND (Self->Opacity IS 255)) Bitmap->Opacity = Self->FirstColourRGB.Alpha + (pos / AlphaDist);
         Action(MT_DrawRGBLine, Bitmap, &line);
         pos += 1024;

         if (line.YCoord IS y) line.XCoord++;
         else line.YCoord--;

         if (line.XEnd IS (x + width-1)) line.YEnd--;
         else line.XEnd++;
      }

      Bitmap->Opacity = opacity;
   }
}

/*****************************************************************************
** Gradient: Free
*/

static ERROR GRADIENT_Free(struct Gradient *Self, APTR Void)
{
   OBJECTPTR drawable;

   if (Self->DrawableID) {
      if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, drawable, Self, NULL);
         ReleaseObject(drawable);
      }
   }

   if (Self->Frames) {
      FreeMemory(Self->Frames);
      Self->Frames = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Gradient: Hide
*/

static ERROR GRADIENT_Hide(struct Gradient *Self, APTR Void)
{
   if (Self->Flags & GRF_VISIBLE) {
      Self->Flags &= ~GRF_VISIBLE;
      if (Self->Head.ObjectFlags & NF_INITIALISED) ActionMsg(AC_Draw, Self->DrawableID, NULL);
      return(ERR_Okay);
   }
   else return(ERR_Okay|ERF_Notified);
}

/*****************************************************************************
** Gradient: Init
*/

static ERROR GRADIENT_Init(struct Gradient *Self, APTR Void)
{
   OBJECTPTR drawable;
   OBJECTID owner_id;
   struct Variable value;

   /*** Find the Drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (!owner_id) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = owner_id;
   }

   /* If dimension settings are missing (e.g. if it is impossible to
   ** determine width, height or a coordinate), then we set the missing
   ** fields to maximum possible values.
   */

   value.Type = FD_DOUBLE;
   value.Double = 0;
   if (!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_X(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_Y(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_XOffset(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_YOffset(Self, &value);
   }

   if (AccessObject(Self->DrawableID, 3000, &drawable) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, drawable, Self, &DrawGradient);
      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Gradient: NewObject
*/

static ERROR GRADIENT_NewObject(struct Gradient *Self, APTR Void)
{
   Self->FirstColourRGB.Red   = 255;
   Self->FirstColourRGB.Blue  = 0;
   Self->FirstColourRGB.Green = 0;
   Self->FirstColourRGB.Alpha = 255;
   Self->LastColourRGB.Red    = 255;
   Self->LastColourRGB.Blue   = 255;
   Self->LastColourRGB.Green  = 255;
   Self->LastColourRGB.Alpha  = 255;
   Self->Direction = GDI_VERTICAL;
   Self->Opacity = 255;
   return(ERR_Okay);
}

/*****************************************************************************
** Gradient: Show
*/

static ERROR GRADIENT_Show(struct Gradient *Self, APTR Void)
{
   if (!(Self->Flags & GRF_VISIBLE)) {
      Self->Flags |= GRF_VISIBLE;
      if (Self->Head.ObjectFlags & NF_INITIALISED) ActionMsg(AC_Draw, Self->DrawableID, NULL);
      return(ERR_Okay);
   }
   else return(ERR_Okay|ERF_Notified);
}

/*****************************************************************************
** Function: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

LONG ReadByte(STRING Str)
{
   LONG Value;
   if (!*Str) return(0);

   if ((*Str >= '0') AND (*Str <= '9')) Value = (*Str - '0') * 16;
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value = ((*Str - 'A')+10) * 16;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value = ((*Str - 'a')+10) * 16;
   else Value = 0;

   Str++;
   if ((*Str >= '0') AND (*Str <= '9')) Value += (*Str - '0');
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value += (*Str - 'A') + 10;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value += (*Str - 'a') + 10;

   return(Value);
}

/*****************************************************************************
** Internal: ReadColour()
** Short:    Reads colours in hex and RGB formats.
*/

static ERROR ReadColour(STRING Colour, LONG *Red, LONG *Green, LONG *Blue, LONG *Alpha)
{
   if (*Colour IS '#') {
      Colour++;
      *Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Blue  = ReadByte(Colour);

      if (Alpha) {
         if (*Colour) Colour++; if (*Colour) Colour++;
         if (*Colour > 0x20) *Alpha = ReadByte(Colour);
      }
   }
   else {
      *Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Blue = StrToInt(Colour);

      if (Alpha) {
         while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
         while ((*Colour) AND ((*Colour < '0') OR (*Colour > '9'))) Colour++;
         if ((*Colour >= '0') AND (*Colour <= '9')) *Alpha = StrToInt(Colour);
         else *Alpha = 255;
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Direction
Short:  Sets the direction of the Gradient colours.
Type:   INDEX
Status: Read/Write

###DESCRIPTION###
<p>By default the Gradient will be drawn vertically, but you may want to have
it drawn horizontally or diagonally.  You will need to set this field to
determine the exact orientation of the gradient.  Valid settings are:</p>

<blockquote>
VERTICAL<br>
HORIZONTAL<br>
TOPDIAGONAL<br>
BOTTOMDIAGONAL
</blockquote>

<p>For further clarification, the TOPDIAGONAL option will draw the
Gradient from the top left down to the bottom right.  The BOTTOMDIAGONAL
option will draw the Gradient from the bottom left up to the top right.
Vertical gradients start from the top down, and horizontal gradients
start from the left to the right.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   FirstColour
Short:  String-based field for setting the first colour of the Gradient.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>You need to set the first colour of the Gradient by writing a string-based
colour to this field.  The colour must be in hexadecimal or separated-decimal
format - for example to create a pure red colour, a setting of '#ff0000' or
'255,0,0' would be valid.</p>

<p>You may also set the alpha channel in the fourth value if you wish
to use gradual alpha blending in your gradient.  The maximum decimal value
for attaining full opacity is 255.</p>

###SEE ALSO###
Fields: FirstColourRGB, LastColour

###END###

*****************************************************************************/

static ERROR SET_FirstColour(struct Gradient *Self, STRING FirstColour)
{
   ReadColour(FirstColour, &Self->FirstColourRGB.Red, &Self->FirstColourRGB.Green, &Self->FirstColourRGB.Blue, &Self->FirstColourRGB.Alpha);
   if (Self->Head.ObjectFlags & NF_INITIALISED) Self->BitsPerPixel = NULL;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   FirstColourRGB
Short:  Defines the first colour of the Gradient in RGB format.
Type:   RGB
Status: Read/Write

###DESCRIPTION###
<p>You can read or write the first colour of the Gradient using this RGB
based field.</p>

###SEE ALSO###
Fields: FirstColour, LastColourRGB

###END###

*****************************************************************************/

static ERROR GET_FirstColourRGB(struct Gradient *Self, struct RGB *RGB)
{
   RGB->Red   = Self->FirstColourRGB.Red;
   RGB->Green = Self->FirstColourRGB.Green;
   RGB->Blue  = Self->FirstColourRGB.Blue;
   return(ERR_Okay);
}

static ERROR SET_FirstColourRGB(struct Gradient *Self, struct RGB *RGB)
{
   Self->FirstColourRGB.Red   = RGB->Red;
   Self->FirstColourRGB.Green = RGB->Green;
   Self->FirstColourRGB.Blue  = RGB->Blue;
   if (Self->Head.ObjectFlags & NF_INITIALISED) Self->BitsPerPixel = NULL;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Forces the Gradient to be drawn to a specific frame.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If this field is set to a valid frame number, the Gradient will only
be drawn when the frame of the container matches the Frame number in this
field.  When set to 0 (the default), the Gradient will be drawn regardless
of the container's frame number.</p>

###SEE ALSO###
Class: Render.Frame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frames
Short:  Determines a sequence of frames that a gradient graphic may be drawn to.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>For long animation sequences you can set the Frames field to determine
individual frames that a gradient graphic should be drawn to.  For instance,
if a drawable has been assigned 20 animation frames and you want a gradient
graphic to appear in frames 5, 6, 7, 8, 17 and 19, you would use this
frame specification:</p>

<pre>
   5-8,17,19
</pre>

<p>Individual frame sets are separated by any non-numeric character such as
a comma, and ranges can be specified through the '-' character.</p>

<p>Setting the Frames field will override any setting that you may have
previously set in the Frame field.</p>

###END###

*****************************************************************************/

static ERROR SET_Frames(struct Gradient *Self, STRING Value)
{
   STRING str;
   LONG frame, end;

   if ((!Value) OR (!(*Value))) return(ERR_Args);

   /*** Calculate the size of the frame array that we are going to need ***/

   Self->TotalFrames = 0;
   str = Value;
   while (*str) {
      if ((*str >= '0') AND (*str <= '9')) {
         frame = StrToInt(str);
         while ((*str >= '0') AND (*str <= '9')) str++;
         if (frame > Self->TotalFrames) Self->TotalFrames = frame;
      }
      else str++;
   }

   Self->TotalFrames++; /* Add an extra frame because we start from #1, not zero */

   /*** Allocate the frame array and record the sequence ***/

   if (AllocMemory((Self->TotalFrames+1) * sizeof(WORD), MEM_DATA, &Self->Frames, NULL) IS ERR_Okay) {
      str = Value;
      while (*str) {
         if ((*str >= '0') AND (*str <= '9')) {
            if ((frame = StrToInt(str)) > 0) {
               while ((*str >= '0') AND (*str <= '9')) str++;
               while ((*str) AND (*str <= 0x20)) str++;
               if (*str IS '-') {
                  while ((*str) AND ((*str < '0') OR (*str > '9'))) str++;
                  if ((end = StrToInt(str)) > 0) {
                     while (frame != end) {
                        Self->Frames[frame] = 1;
                        if (frame < end) frame++;
                        else frame--;
                     }
                     Self->Frames[frame] = 1;
                  }
                  while ((*str >= '0') AND (*str <= '9')) str++;
               }
               else Self->Frames[frame] = 1;
            }
            else while ((*str >= '0') AND (*str <= '9')) str++;
         }
         else str++;
      }

      Self->Frames[Self->TotalFrames] = -1; /* Sequence termination */
   }
   else return(ERR_AllocMemory);

   Self->Frame = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   LastColour
Short:  String-based field for setting the last colour of the Gradient.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>You need to set the last colour of the Gradient by writing a string-based
colour to this field.  The colour must be in hexadecimal or separated-decimal
format - for example to create a pure red colour, a setting of '#ff0000' or
'255,0,0' would be valid.</p>

<p>You may also set the alpha channel in the fourth value if you wish
to use gradual alpha blending in your gradient.  The maximum decimal value
for attaining full opacity is 255.</p>

###SEE ALSO###
Fields: LastColourRGB, FirstColour

###END###

*****************************************************************************/

static ERROR SET_LastColour(struct Gradient *Self, STRING LastColour)
{
   ReadColour(LastColour, &Self->LastColourRGB.Red, &Self->LastColourRGB.Green, &Self->LastColourRGB.Blue, &Self->LastColourRGB.Alpha);
   if (Self->Head.ObjectFlags & NF_INITIALISED) Self->BitsPerPixel = NULL;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   LastColouRGB
Short:  Defines the last colour of the Gradient in RGB format.
Type:   RGB
Status: Read/Write

###DESCRIPTION###
<p>You can read or write the last colour of the Gradient using this RGB
based field.</p>

###SEE ALSO###
Fields: LastColour, FirstColourRGB

###END###

*****************************************************************************/

static ERROR GET_LastColourRGB(struct Gradient *Self, struct RGB *RGB)
{
   RGB->Red   = Self->LastColourRGB.Red;
   RGB->Green = Self->LastColourRGB.Green;
   RGB->Blue  = Self->LastColourRGB.Blue;
   return(ERR_Okay);
}

static ERROR SET_LastColourRGB(struct Gradient *Self, struct RGB *RGB)
{
   Self->LastColourRGB.Red   = RGB->Red;
   Self->LastColourRGB.Green = RGB->Green;
   Self->LastColourRGB.Blue  = RGB->Blue;
   if (Self->Head.ObjectFlags & NF_INITIALISED) Self->BitsPerPixel = NULL;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a gradient.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A gradient can be given a fixed or relative height by setting this field
to the desired value.  To set a relative height, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct Gradient *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerheight, ycoord;
   DOUBLE value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_LONG, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Dimensions & DMF_FIXED_HEIGHT) value = Self->Height;
      else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         value = Self->Height * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_Y) ycoord = Self->YCoord;
         else ycoord = (containerheight * Self->YCoord) / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - ycoord - Self->YOffset;
         else value = containerheight - ycoord - ((containerheight * Self->YOffset) / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetHeight, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct Gradient *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Gradient] A height of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Gradient] A height of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetHeight, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Opacity
Short:  Determines the level of translucency applied to a gradient.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>This field determines the translucency level of a gradient.  The default
setting is 100%, which means that the gradient will be solid.  Any other
value that you set here will alter the impact of a gradient over the
destination drawable.  High values will retain the boldness of the gradient,
while low values can render it close to invisible.</p>

<p>Please note that the use of translucency will always have an impact on the
time it normally takes to draw a gradient.  The use of translucency also
requires that the drawable area is buffered, as read access is required to
perform the blending algorithm.</p>

###END###

****************************************************************************/

static ERROR GET_Opacity(struct Gradient *Self, FLOAT *Value)
{
   *Value = Self->Opacity * 100 / 255;
   return(ERR_Okay);
}

static ERROR SET_Opacity(struct Gradient *Self, FLOAT *Value)
{
   if (*Value < 0) Self->Opacity = 0;
   else if (*Value > 100) Self->Opacity = 255;
   else Self->Opacity = *Value * 255 / 100;

   if (Self->Head.ObjectFlags & NF_INITIALISED) Self->BitsPerPixel = NULL;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a gradient.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A gradient can be given a fixed or relative width by setting this field
to the desired value.  To set a relative width, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct Gradient *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerwidth;
   DOUBLE xcoord, value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_LONG, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Dimensions & DMF_FIXED_WIDTH) value = Self->Width;
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         value = Self->Width * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_X) xcoord = Self->XCoord;
         else xcoord = containerwidth * Self->XCoord / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - xcoord - Self->XOffset;
         else value = containerwidth - xcoord - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetWidth, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Width(struct Gradient *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Gradient] A width of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Gradient] A width of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetWidth, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a gradient.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a gradient can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Gradient *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerwidth;
   DOUBLE width, value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_LONG, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_X) value = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         value = Self->XCoord * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - width - Self->XOffset;
         else value = containerwidth - width - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetX, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetX, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Gradient *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetX, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a gradient.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the gradient will be
drawn from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the Gradient is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the gradient will be drawn at an X coordinate calculated from the formula
&quot;XCoord = ContainerWidth - GradientWidth - XOffset&quot;.</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Gradient *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerwidth;
   DOUBLE width, value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_LONG, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_XOFFSET) value = Self->XOffset;
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->XOffset * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_X) value = containerwidth - (Self->XCoord + width);
         else value = containerwidth - ((Self->XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetXOffset, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Gradient *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->XOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetXOffset, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of a gradient.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a Gradient can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct Gradient *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerheight, height;
   DOUBLE value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_LONG, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_Y) value = Self->YCoord;
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         value = Self->YCoord * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = (containerheight * Self->Height) / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - height - Self->YOffset;
         else value = containerheight - height - ((containerheight * Self->YOffset) / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetY, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Gradient *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetY, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a gradient.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the gradient will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the gradient is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the gradient will be drawn at a Y coordinate calculated from the formula
&quot;YCoord = ContainerHeight - GradientHeight - YOffset&quot;.</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Gradient *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerheight;
   DOUBLE value, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_LONG, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_YOFFSET) value = Self->YOffset;
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->YOffset * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_Y) value = containerheight - (Self->YCoord + height);
         else value = containerheight - ((Self->YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetYOffset, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Gradient *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetYOffset, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

                  "Won't you join me on the perennial quest
                   Reaching into the dark, retrieving light
                   Search for answers on the perennial quest
                   Where dreams are followed, and time is a test"

                        - Death

*****************************************************************************/
